let _ = Mltop.add_known_module "coq-elpi.cs"

# 3 "apps/cs/src/coq_elpi_cs_hook.mlg"
 

open Elpi
open Elpi_plugin
open Coq_elpi_arg_syntax
open Coq_elpi_vernacular

let elpi_cs_hook program env sigma ((proji, u), params1, c1) (t2, args2) =
	  let atts = [] in
     let rhs = Reductionops.Stack.zip sigma (t2, Reductionops.Stack.append_app_list args2 Reductionops.Stack.empty) in
     let sigma, (goal_ty, _) = Evarutil.new_type_evar env sigma Evd.UnivRigid in
	  let sigma, goal = Evarutil.new_evar env sigma goal_ty in
	  let goal_evar, _ = EConstr.destEvar sigma goal in
    let nparams = Structures.Structure.projection_nparams proji in
	  let query state =
         let loc = Elpi.API.State.get Coq_elpi_builtins_synterp.invocation_site_loc state in
		 let state, q, gls =
      let lhs = match params1 with
        | Some params1 -> EConstr.mkApp (EConstr.mkConstU (proji, u), Array.of_list params1)
        | None -> EConstr.mkConstU (proji, u) in
			Coq_elpi_HOAS.solvegoals2query sigma [goal_evar] loc ~main:[lhs; rhs]
			  ~in_elpi_tac_arg:Coq_elpi_arg_HOAS.in_elpi_tac_econstr ~depth:0 ~base:() state in
		 let state, qatts = atts2impl loc Summary.Stage.Interp ~depth:0 state atts q in
		 let state = API.State.set Coq_elpi_builtins.tactic_mode state true in
		 state, qatts, gls
	  in
	  let loc = Loc.initial Loc.ToplevelInput in
      match Interp.get_and_compile ~loc program with
	  | None -> None
	  | Some (cprogram, _) ->
		 begin match Interp.run ~loc cprogram (Fun query) with
		 | API.Execute.Success solution ->
			let gls = Evar.Set.singleton goal_evar in
			let sigma, _, _ = Coq_elpi_HOAS.solution2evd ~eta_contract_solution:false sigma solution gls in
			if Evd.is_defined sigma goal_evar then
           let constant = EConstr.to_constr sigma goal in
			  let args_goal, t = Reduction.whd_decompose_lambda env constant in
			  let args_type = List.rev_map Context.Rel.Declaration.get_type args_goal in
			  let args_type = List.map EConstr.of_constr args_type in
           let args = snd (Constr.decompose_app_list t) in
           let params, projs = CList.chop nparams args in
           let i = Structures.Structure.projection_number env proji in
           let lhs = List.nth projs i in
			  let lhs = EConstr.of_constr lhs in
			  let _, sk1 = EConstr.decompose_app sigma lhs in
			  let open Structures.CanonicalSolution in
			  Some (sigma, {
				 constant = EConstr.of_constr constant;
				 abstractions_ty = args_type;
				 body = lhs;
				 nparams;
				 params = List.map EConstr.of_constr params;
				 cvalue_abstraction = None;
				 cvalue_arguments = Array.to_list sk1})
			else None
		 | API.Execute.NoMoreSteps
		 | API.Execute.Failure -> None
		 end

[%%if coq = "8.20"]
let adapt_hook f : Evarconv.hook = fun env sigma (s,l,t) x -> f env sigma (s,Some l,t) x
[%%else]
let adapt_hook f : Evarconv.hook = f
[%%endif]

let add_cs_hook =
  let cs_hook_program = Summary.ref ~name:"elpi-cs" None in
  let cs_hook env sigma proj pat =
    match !cs_hook_program with
    | None -> None
    | Some h -> elpi_cs_hook h env sigma proj pat in
  let name = "elpi-cs" in
  Evarconv.register_hook ~name (adapt_hook cs_hook);
  let inCs =
    let cache program =
      cs_hook_program := Some program;
      Evarconv.activate_hook ~name in
    let open Libobject in
    declare_object
    @@ superglobal_object_nodischarge "ELPI-CS" ~cache ~subst:None in
  fun program -> Lib.add_leaf (inCs program)



let () = Vernacextend.static_vernac_extend ~plugin:(Some "coq-elpi.cs") ~command:"ElpiCS" ~classifier:(fun _ -> Vernacextend.classify_as_sideeff) ?entry:None 
         [(Vernacextend.TyML (false, Vernacextend.TyTerminal ("Elpi", 
                                     Vernacextend.TyTerminal ("CS", Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_qualified_name), 
                                                                    Vernacextend.TyNil))), 
         (let coqpp_body p atts = Vernactypes.vtdefault (fun () -> 
# 88 "apps/cs/src/coq_elpi_cs_hook.mlg"
                                                                  
     let () = ignore_unknown_attributes atts in
     add_cs_hook (snd p) 
                                  ) in fun p
         ?loc ~atts () -> coqpp_body p (Attributes.parse 
# 88 "apps/cs/src/coq_elpi_cs_hook.mlg"
             any_attribute
          atts)), None))]

